//#################################################################
//	Testbench for the EPT2232H DOCK BOARD SPI TO FPGA 
//
//	File name:	tb_ept_2232h_top.v
//	Author:		Richard Jolly
//	Date:		04/21/2012
//
//	Copyright 	Earth People Technology Inc. 2012
//################################################################

`timescale 1ns / 100ps

`include "../src/define.v"
`include "../src/tb_define.v"

module tb_ept_ft2232h_top;

   //----------------------------------------------
   // Parameter Declarations  
   //----------------------------------------------
    parameter BEGIN_TITLE    = 8'h00;
    parameter WRITE_SECTION  = 8'h01;
    parameter READ_SECTION   = 8'h02;
    parameter UART_READ      = 8'h03;
    parameter UART_WRITE     = 8'h04;

	
//---------------------------------------------------------------------------------------------------------------
//	wire and register declarations
//---------------------------------------------------------------------------------------------------------------
	reg						  reset;

    reg                        clk_10;       // 10 MHzClock
    reg                        clk_50 ;       // 50MHz Clock
    reg                        clk_100;       // 100MHzClock
    reg                        clk_66;        // 66 MHz Clock

   //HSP FT2232H Top
	wire [7:0]                 spi_slave_data_in;
	wire                       spi_slave_wr_en;
	wire                       spi_slave_write_block_en;
	wire                       spi_slave_ack_done;
	wire [7:0]                 spi_slave_data_out;
	
	//SPI Master Top
	reg  [7:0]                 spi_master_din;
	reg                        spi_master_we_in;
	wire                       spi_master_rdy_in;
	wire  [7:0]                spi_master_dout;
	wire                       spi_master_we_out_n;
	reg                        spi_master_rdy_out;
	
	//SPI Bus
	wire                       spi_mosi;
	wire                       spi_clk;
	wire                       spi_miso;
	wire                       spi_cs;
	
	//UART signals
	wire                       uart_rxd;
	wire                       uart_txd;
	
	//Read/Write Registers
	reg [7:0]                  spi_read_value;
	reg [7:0]                  spi_reg_write;
	reg [7:0]                  spi_reg_address;	
	
	//Loop Control
	integer                    idx;
	reg [7:0]                  register_write [2:0];
	reg [7:0]                  num_read_cycles;
	reg                        first_time_thru_loop;

	//Read Control Registers
	wire [7:0]                 read_control_register;
	reg	[7:0]                  read_control_register_index;

	//read and write register bytes
	reg	[7:0]                  host_read_register;
	wire [7:0]                 write_register_value;

	reg							write_to_read_control_register_block;
	reg	[7:0]					write_read_control_word;

	//Write Control Registers
	wire [7:0]                 write_control_register;
	reg	[7:0]                  write_control_register_index;
	reg	[7:0]                  write_to_write_control_register_block;
	reg                        write_control_word;

	//Host control 
	wire [25:0]                host_bus_model_addr;
	wire [15:0]                host_bus_model_data;
	reg [7:0]                  host_write_word;
	reg [7:0]                  host_transmit_word_1;
	reg [15:0]                  host_transmit_num_of_words;
	reg [15:0]                  dut_received_word;
	
	//Test Bench Stimulus Host Registers
	reg	[7:0]                  host_register;
	reg	[7:0]                  host_readback;
	reg	[7:0]                  host_status_byte;
	reg                        host_read_status;
	reg                        host_status_we;
	reg  [7:0]                 host_data_in;
	wire [7:0]                 host_data_out;
	reg                        host_read_byte;
	reg                        host_write_byte;
	reg  [7:0]                 host_write_value;

	//DUT pass registers
	wire                       usb_rxf_n;
	wire                       usb_txe_n;
	wire                       usb_rd_n;
	wire                       usb_wr;
	reg                        host_ft245_reset;
	wire                       host_ft245_pwren;
	reg                        host_ft245_test;
	wire [7:0]                  usb_data;
	
	//Status Registers
	reg [7:0]                  serial_uart_status_register;
	reg [7:0]                  serial_uart_status_register_model;
	wire [7:0]                 serial_uart_control_register_model;	
	wire [15:0]                serial_uart_receive_register_model;                
	
	//UART control signals
	reg [15:0]                 serial_uart_send_register_model;
	reg [15:0]                 serial_uart_control_register;
	reg [15:0]                 serial_uart_receive_register;
	reg [15:0]                 serial_uart_send_register;
	
   //-----------------------------------------------
   // Generate 10 MHz Clock with 'cycle ns period
   //-----------------------------------------------
   initial // Clock generator
     begin
	clk_10 = 0; 
	forever 
	  #`CYCLE_10 clk_10 = !clk_10;
     end

	 //-----------------------------------------------
   // Generate 150 MHz Clock with 'cycle ns period
   //-----------------------------------------------
   initial // Clock generator
     begin
	clk_50 = 0; 
	forever 
	  #`CYCLE_50 clk_50 = !clk_50;
     end

   //-----------------------------------------------
   // Generate 100 MHz Clock with 'cycle ns period
   //-----------------------------------------------
   initial // Clock generator
     begin
	clk_100 = 0; 
	forever 
	  #`CYCLE_100 clk_100 = !clk_100;
     end

   //-----------------------------------------------
   // Generate 66 MHz Clock with 'cycle ns period
   //-----------------------------------------------
   initial // Clock generator
     begin
	clk_66 = 0; 
	forever 
	  #`CYCLE_66 clk_66 = !clk_66;
     end

//-------------------------------------------------------------------------------------------------------------
//	Apply Stimulus
//-------------------------------------------------------------------------------------------------------------

	initial
	begin


		call_title(BEGIN_TITLE);
		reset						     = 1'b1;
		host_register                    = 8'h0;
		host_read_status                 = 1'b0;
		host_read_byte                   = 1'b0;
		host_write_byte                  = 1'b0;
		host_write_value                 = 8'h0;
		host_status_we                   = 1'b0;
		host_status_byte                 = 8'h0;
		host_ft245_test                  = 1'b0;
		host_ft245_reset                 = 1'b0;
		serial_uart_status_register		 =	8'h0;
		serial_uart_status_register_model	=	8'h0;
		host_transmit_num_of_words       = 8'h01;
		first_time_thru_loop             = 1'b0;
		host_transmit_word_1             = 8'h85;


		#(100 * `CYCLE)
		reset	=			1'b0;

		#(50 * `CYCLE)
		reset	=			1'b1;




		call_title(WRITE_SECTION);
		repeat(10)
		begin

			#(15000 * `CYCLE )
			//write_outgoing_message_serial_uart_comms_model( host_transmit_word_1, 
			 //                                               host_transmit_num_of_words);

			 //Wait for UART Tiny to receive the byte and transmit
			 //it to the FTDI Host
	         //while(!host_status_we)
			 //begin
	         //    host_check_status(host_status_byte);
			//	 host_status_we = host_status_byte && 8'h01;
	         //end
			 //#(500 * `CYCLE )
             //Read the byte received from SPI Tiny from the Host Model
             //host_bus_read(host_readback);
			 
			 //Compare the byte read using SPI Tiny from the 
			 // SPI model.
             compare_spi_write_read(host_transmit_word_1, host_readback, 8'h00);


			//host_transmit_word_1 = host_transmit_word_1 + 1'd1;
			
		//Read host_status_byte to set host_status_we to zero
	         //while(host_status_we)
			 //begin
	         //    host_check_status(host_status_byte);
			//	 host_status_we = host_status_byte[0] && 1'b1;
	         //end
		

		end // repeat(num_read_cycles)

		 #(5000 * `CYCLE )

        host_write_value = 8'ha5;
		//Read two words from the DUT 
		call_title(READ_SECTION);
		repeat(10)
		begin

		     // Command Host to write a byte to the UART Tiny
             //from the UART model then read the byte from the UART
             // Tiny.			 
		     #(15000 * `CYCLE)
		     host_bus_write( host_write_value);
			 read_incoming_message_serial_uart_comms_model( dut_received_word, 8'h01 );

			 #(500 * `CYCLE )
			 compare_host_read( host_write_word, dut_received_word, num_read_cycles );
		
		     //	num_read_cycles = num_read_cycles + 1;
		     host_write_value = host_write_value + 1'd1;
		     //end // repeat(num_read_cycles)
		
		     #(1000 * `CYCLE);
		end //repeat()



	  #(20000 * `CYCLE);
	stopsim;


	end	//stimulus
//--------------------------------------------------
//	Task: call title
//--------------------------------------------------
   task call_title;
	input [7:0] select_title;
      begin
		case ( select_title ) 
		8'h00:
		$display("\n\n\n\nThis testbench will test the UART Communications at time %0t####\n",$time);
		8'h01:
		$display("\n\n\n\nWrite to the UART at time %0t####\n",$time);
		8'h02:
		$display("\n\n\n\nRead from the UART at time %0t####\n",$time);
		8'h03:
		$display("\n\n\n\nRead Section of Testbench The Control Command at time %0t####\n",$time);
		default:
		$display("\n\n\n\nTest bench running.... at time %0t####\n",$time);
		endcase

      end
   endtask // host_bus_read

//--------------------------------------------------
//	Task: write_outgoing_message_serial_uart_comms_model
//	For Host Model only
//	Write two words to the serial_iic_comms,
//	when done, start the transmit process.
//--------------------------------------------------
   task	write_outgoing_message_serial_uart_comms_model;
         input		[7:0]	serial_byte_out;
	     input		[7:0]		num_of_words;

	     reg				first_time_thru_loop;
    begin

	//first_time_thru_loop = 1'b0;

	while( num_of_words ) 
	begin

			serial_uart_send_register_model = serial_byte_out;
	 		#(200 * `CYCLE);

			serial_uart_status_register_model[1] = 1'b1;
	 		#(200 * `CYCLE);

			wait( serial_uart_control_register_model[2] );
			#(200 * `CYCLE);

			serial_uart_status_register_model[1] = 1'b0;
	 		#(200 * `CYCLE);

			num_of_words = num_of_words - 1;
			//first_time_thru_loop <= first_time_thru_loop + 1;
			#(200 * `CYCLE);

	end

    end
   endtask 

//--------------------------------------------------
//	Task: read_incoming_message_serial_uart_comms
//	For Host Model Only
//	Once the serial_iic_comms receives a message,
//	read available words.
//--------------------------------------------------
   task	read_incoming_message_serial_uart_comms;
      output	[15:0]	received_data_in;
	input				read_received_word;
      begin
	
	if( serial_uart_control_register[3] )
	begin
		received_data_in = serial_uart_receive_register;
		 #(20 * `CYCLE);

		serial_uart_status_register[3] = 1'b1;
		 #(20 * `CYCLE);

		serial_uart_status_register[3] = 1'b0;
		#(20 * `CYCLE);

	end
	else 
		$display("Failded to read Receive UART from DUT. \nserial_uart_control_register[3] .\n");

      end
   endtask 

//--------------------------------------------------
//	Task: write_outgoing_message_serial_uart_comms
//	For Host Model only
//	Write two words to the serial_iic_comms,
//	when done, start the transmit process.
//--------------------------------------------------
   task	write_outgoing_message_serial_uart_comms;
      input		[15:0]	transmit_data_out_1;
	input		[7:0]		num_of_words;

	reg				first_time_thru_loop;



      begin

	first_time_thru_loop = 1'b0;

	while( num_of_words ) 
	begin

			serial_uart_send_register = transmit_data_out_1;
	 		#(200 * `CYCLE);

			serial_uart_status_register[1] = 1'b1;
	 		#(200 * `CYCLE);

			wait( serial_uart_control_register[2] );
			#(200 * `CYCLE);

			serial_uart_status_register[1] = 1'b0;
	 		#(200 * `CYCLE);

			num_of_words = num_of_words - 1;
			first_time_thru_loop <= first_time_thru_loop + 1;
			#(200 * `CYCLE);

	end

      end
   endtask 
//--------------------------------------------------
//	Task: read_incoming_message_serial_uart_comms_model
//	For Host Model Only
//	Once the serial_iic_comms receives a message,
//	read available words.
//--------------------------------------------------
   task	read_incoming_message_serial_uart_comms_model;
      output	[15:0]	received_data_in;
	  input				read_received_word;
	  reg [15:0]     timeout_cntr;
      begin
	     timeout_cntr = 0;
	     wait( serial_uart_control_register_model[3] );
	         //if(timeout_cntr < 16'h5000)	 
	        // begin
		         received_data_in = serial_uart_receive_register_model;
		         #(20 * `CYCLE);

		         serial_uart_status_register_model[3] = 1'b1;
		         #(20 * `CYCLE);

		         serial_uart_status_register_model[3] = 1'b0;
		         #(20 * `CYCLE);

	        // end
	        // else 
		    //     $display("Failded to read Receive UART from DUT. \n
			//	 serial_uart_control_register[3] .\n");

      end
   endtask 

//--------------------------------------------------
//	Task: SPI read, Register and Value
//--------------------------------------------------
   task spi_bus_read;
   input [7:0] read_register_address;
   output [7:0] read_register_value;
   begin
	     spi_master_din = read_register_address;
	     spi_master_we_in = 1'b1;
	     #(50 * `CYCLE);
	     wait (spi_master_rdy_in);
	     #(50 * `CYCLE);
	     spi_master_we_in = 1'b0;
	     wait (!spi_master_rdy_in);
	     #(200 * `CYCLE);
	     spi_master_rdy_out = 1'b1;
	     wait(!spi_master_we_out_n);
	     read_register_value = spi_master_dout;
	     #(200 * `CYCLE);
	     spi_master_rdy_out = 1'b0;
	     wait(spi_master_we_out_n);
    end
   endtask // host_bus_read
//--------------------------------------------------
//	Task: SPI read, Value only
//--------------------------------------------------
   task spi_read_byte;
    output [7:0] read_register_value;
    begin
	     #(50 * `CYCLE);
	     spi_master_rdy_out = 1'b1;
	     wait(!spi_master_we_out_n);
	     read_register_value = spi_master_dout;
	     #(200 * `CYCLE);
	     spi_master_rdy_out = 1'b0;
	     wait(spi_master_we_out_n);
    end
   endtask // host_bus_read


//--------------------------------------------------
//	Task: SPI write, Register and Value
//--------------------------------------------------
   task spi_bus_write;
   input [7:0] write_register_address;
   input [7:0] write_register_value;
   begin
	     spi_master_din = write_register_address;
	     spi_master_we_in = 1'b1;
	     #(50 * `CYCLE);
	     wait (spi_master_rdy_in);
	     #(50 * `CYCLE);
	     spi_master_we_in = 1'b0;
	     wait (!spi_master_rdy_in);
	     #(200 * `CYCLE);
	     spi_master_din = write_register_value;
	     spi_master_we_in = 1'b1;
	     #(50 * `CYCLE);
	     wait (spi_master_rdy_in);
	     #(50 * `CYCLE);
	     spi_master_we_in = 1'b0;
	     wait (!spi_master_rdy_in);
   end
   endtask // host_bus_write

   //--------------------------------------------------
//	Task: SPI write Value only
//--------------------------------------------------
   task spi_write_byte;
   input [7:0] write_register_value;
   begin
	     spi_master_din = write_register_value;
	     spi_master_we_in = 1'b1;
	     #(50 * `CYCLE);
	     wait (spi_master_rdy_in);
	     #(50 * `CYCLE);
	     spi_master_we_in = 1'b0;
	     wait (!spi_master_rdy_in);
	     #(200 * `CYCLE);
   end
   endtask // host_bus_write

//--------------------------------------------------
//	Task: compare read register from configuration
//--------------------------------------------------
   task compare_spi_write_read;
	input [7:0] register_write;
	input [7:0] register_read;
	input [7:0] register_address;
	reg	start_compare_read_cycle;
	begin
		start_compare_read_cycle = 1'b1;
	 	#(20 * `CYCLE);
		if( register_read[7:0] != register_write[7:0] )
		     $display("Compare FAILED at register %h\n", register_address);
		 else
		     $display("Compare PASSED at register %h\n", register_address);
		start_compare_read_cycle = 1'b0;
		 #(20 * `CYCLE);
		
	end
    endtask

//--------------------------------------------------
//	Task: Host read register memory
//--------------------------------------------------
   task host_bus_read;
    output [7:0] read_register_value;
    begin
	     host_read_byte = 1'b1;
	     #(20 * `CYCLE);
	     read_register_value = host_data_out;
	     #(20 * `CYCLE);
	     host_read_byte = 1'b0;
	     #(20 * `CYCLE);
    end
   endtask // host_bus_read


//--------------------------------------------------
//	Task: Host write register memory
//--------------------------------------------------
   task host_bus_write;
   input [7:0] write_register_value;
   begin
	     #(20 * `CYCLE);
         host_write_byte = 1'b1;
	     host_data_in = write_register_value;
	     #(20 * `CYCLE);
	     host_write_byte = 1'b0;
	     #(20 * `CYCLE);
   end
   endtask // host_bus_write

//--------------------------------------------------
//	Task: compare read register 
//--------------------------------------------------
   task compare_host_read;
	input [7:0]	write_register;
	input [7:0]	read_register;
	input	[7:0]		read_cycle_number;
	reg	start_compare_read_cycle;


	begin
		start_compare_read_cycle = 1'b0;


		$display("Performing the Compare of read and write words\n");
		start_compare_read_cycle = 1'b1;
	 	#(200* `CYCLE);
		if( write_register[7:0] != read_register[7:0] )
			$display("Compare failed at read cycle %h\n", read_cycle_number);
		else
			$display("Compare Succeeded!!! at read cycle %h\n", read_cycle_number);
		start_compare_read_cycle = 1'b0;
		 #(200 * `CYCLE);
		
	end
    endtask

//--------------------------------------------------
//	Task: Check Host Status
//--------------------------------------------------
   task host_check_status;
    output [7:0] host_status_byte;
    begin
	     host_read_status = 1'b1;
	     #(20 * `CYCLE);
		 host_status_byte = host_data_out;
	     host_read_status = 1'b0;
	     #(20 * `CYCLE);
    end
   endtask // host_bus_write

   //-----------------------------------------------
   // Task: stopsim
   //-----------------------------------------------
   task stopsim;
      begin
   	 $display("Simulation Finished at time %0t\n",$time);
	 $stop;
      end
   endtask // stopsim

  
   //-----------------------------------------------
   // Instantiate HSP FT2232 Top DUT
   //-----------------------------------------------
   ept_usb_development_top DUT
    (
	.CLK                       (clk_66),
	.RST                       (reset),

	.UART_OUT                  (uart_txd),
	.UART_IN                   (uart_rxd),

	.FTDI_RXF                  (usb_rxf_n),
	.FTDI_TXE                  (usb_txe_n),
	.FTDI_DATA                 (usb_data),

	.FTDI_RD                   (usb_rd_n),
	.FTDI_WR                   (usb_wr),
	.FTDI_SI                   ()
	);

   //-----------------------------------------------
   // Instantiate Host UART Model
   //-----------------------------------------------
   serial_uart_comms_model 					serial_uart_comms_model_inst
     (
		.CLK                   (clk_66),
		.RESET                 (reset),

		.UART_RXD_IN           (uart_txd),
		.UART_TXD_OUT          (uart_rxd),

		.CONTROL_REGISTER      (serial_uart_control_register_model),
		.STATUS_REGISTER       (serial_uart_status_register_model),
		.RECEIVE_REGISTER      (serial_uart_receive_register_model), 
		.SEND_REGISTER         (serial_uart_send_register_model )
	
 
      );   

   //-----------------------------------------------
   // Instantiate Host FT Model
   //-----------------------------------------------
   host_ft245_model            HOST_BUS_MODEL
  (
   	.CLK                       (clk_66),
   	.RST_N                     (reset),

	.HOST_FT245_RXF_N          (usb_rxf_n),
	.HOST_FT245_TXE_N          (usb_txe_n),
	.HOST_FT245_RD_N           (usb_rd_n),
	.HOST_FT245_WR             (usb_wr),
	.HOST_FT245_RESET_N        (host_ft245_reset),
	.HOST_FT245_PWREN_N        (host_ft245_pwren),
	.HOST_FT245_TEST           (host_ft245_test),
	.HOST_FT245_DATA           (usb_data),

	.DATA_IN                   (host_data_in),
	.DATA_OUT                  (host_data_out),
	.READ_FT_245               (host_write_byte),
	.WRITE_FT_245              (host_read_byte),
	.READ_FT_STATUS            (host_read_status)
   );


endmodule // tb_hsp_ft2232_top.v


